/*
 * Copyright (c) 2008  Yuri Chernyavsky
 * Copyright (c) 2007  Gregor Richards
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#include <libgen.h>
#include <signal.h>
#include <sys/wait.h>
#include <unistd.h>
#include <limits.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <libglademm.h>
#include <glibmm.h>

#include <sigc++/class_slot.h>
#include <gtkmm/frame.h>
#include <gtkmm/window.h>
#include <gtkmm/box.h>
#include <gtkmm/button.h>
#include <gtkmm/label.h>
#include <gtkmm/progressbar.h>
#include <gtkmm/filechooserdialog.h>
#include <gtkmm/main.h>
#include <gtkmm/entry.h>
#include <gtkmm/filechooserdialog.h>
#include <gtkmm/stock.h>
#include <gtkmm/image.h>
#include <gtkmm/comboboxentrytext.h>
#include <gtkmm/comboboxtext.h>
#include <gtkmm/scale.h>
#include <gtkmm/buttonbox.h>
#include <gtkmm/spinbutton.h>
#include <gtkmm/liststore.h>
#include <gtkmm/table.h>

#include <queue>
#include <vector>
#include <map>
#include <iostream>
#include <fstream>
#include <cstring>
#include <sstream>
#include <cstdlib>
#include <cstdio>

#include <locale.h>
#include "i18n.h"

#include "smv_encode.h"
#include "smv_gui.h"
#include "i18n.h"

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

static char *force_input_file = NULL;

// Encoder invocation
const std::string backend_exec("smv_encode -b");
const std::string homedir(getenv("HOME"));
// Configuration file name
const std::string config_file(homedir + "/.smv_gui.conf");

// Read exactly len bytes. 
// Return 1 upon success, 0 upon failure
int myread(int fd, char* buf, int len) {
  int count;
  char *ptr = buf;

  while(len > 0) {
    if ((count = read(fd, ptr, len)) <= 0) {
      return 0;
    }
    len -= count;
    ptr += count;
  }
  
  return 1;
}

// Make the child process group leader
void Backend::child_setup() {
  setpgid(0,0);
}

// Cancel the encoding
void Backend::cancel() {
  kill(-child_pid, SIGINT);
}

// Brutally kill
void Backend::stop() {
   kill(-child_pid, SIGKILL);
}

// Read an encoder message
// Note that we can assume that the
// queue is non-empty as we access it
// only upon an appropriate signal
msg_t Backend::get_message() {
  Glib::Mutex::Lock lock (message_queue_mutex_);
  msg_t result = message_queue_.front();
  message_queue_.pop();
  return result;
}

// Push a message to message queue
// The queue is protected by mutex
void Backend::push_message(msg_t msg) {
  Glib::Mutex::Lock lock (message_queue_mutex_);
  message_queue_.push(msg);
}

Backend::Backend(Frontend* frontend) : frontend_(frontend) {}

Backend::~Backend() {}

// Reap the encoder
int Backend::finish() {
  int status, ret;

  ret = waitpid(child_pid, &status, WUNTRACED);

  Glib::spawn_close_pid(child_pid);

  if (WIFEXITED(status)) {
    return (WEXITSTATUS(status) == 0);
  }
  return 0;
}

// Launch the encoder process
void Backend::launch(const char* inp, const char* out, const
		    std::vector<Parameter*>& params)
{
  std::vector<std::string> argv;
  std::stringstream cmd;

  argv.push_back("/bin/sh");
  argv.push_back("-c");

  cmd << backend_exec << " ";
  for (std::vector<Parameter*>::const_iterator i = params.begin();
       i != params.end(); i++) {
    cmd << (*i)->get_flag_and_val() << " ";
  }
  cmd << "\"" << inp << "\"" << " ";
  cmd << "\"" << out << "\"" << " ";

  argv.push_back(cmd.str());

  printf("DEBUG running %s\n", cmd.str().c_str());

  Glib::spawn_async_with_pipes(".", argv, 
			       Glib::SpawnFlags(Glib::SPAWN_SEARCH_PATH | 
						Glib::SPAWN_DO_NOT_REAP_CHILD), 
			       mem_fun(*this, &Backend::child_setup), 
			       &child_pid, 0, &child_fd, 0);
  
  // Handle messages in a separate thread
  Glib::Thread::create(sigc::mem_fun(*this, &Backend::thread_function), false);
}

// Message handler thread
void Backend::thread_function()
{
  int i;
  msg_t msg;
  
  // Notify the main thread about encoding start, 
  // encoding end and each message
  frontend_->start_dispatcher()();
  while (myread(child_fd, (char*)&msg, sizeof(msg_t))) {
    push_message(msg);
    frontend_->message_dispatcher()();
  }
  frontend_->end_dispatcher()();
}

// Set Convert button
void Frontend::set_convert_button() {
  convert_button_connection_.disconnect();
  convert_button_->set_image(*convert_img_);
  convert_button_->set_label(i18n("Convert"));
  convert_button_connection_ = convert_button_->signal_clicked().connect(
       sigc::mem_fun(*this, &Frontend::on_convert));
}

// Set cancel button
void Frontend::set_cancel_button() {
  convert_button_connection_.disconnect();
  convert_button_->set_image(*cancel_img_);
  convert_button_->set_label(i18n("Cancel"));
  convert_button_connection_ = convert_button_->signal_clicked().connect(
         sigc::mem_fun(*this, &Frontend::on_cancel));
}

// Update gui after init
void Frontend::update_gui_after_init() {
  if (force_input_file) {
    input_file_name_->set_sensitive(0);
    input_file_name_->set_text(force_input_file);
    input_file_chooser_button_->set_sensitive(0);
    update_output_filename(force_input_file);
  }
}

// Update gui upon start encoding
void Frontend::update_gui_upon_start_encoding() {
  set_cancel_button();
  convert_button_->set_sensitive(0);
  progress_bar_->set_text(" ");

  // Can we do it in one command by putting all controls in one event box?
  input_file_name_->set_sensitive(0);
  output_file_name_->set_sensitive(0);
  input_file_chooser_button_->set_sensitive(0);
  output_file_chooser_button_->set_sensitive(0);
  save_param_button_->set_sensitive(0);
  default_param_button_->set_sensitive(0);
  for (std::vector<Parameter*>::const_iterator i = parameters_.begin();
       i != parameters_.end(); i++) {
    (*i)->get_widget()->set_sensitive(0);
  }
}

// Update gui upon end encoding
void Frontend::update_gui_upon_end_encoding(int encoder_ok, int success) {
  if (!encoder_ok) {
    progress_bar_->set_text(i18n("Failed to run backend"));
  } else {
    
    // If the encoding was successful, clear
    // the input and output file names
    if (success && !force_input_file) {
      input_file_name_->set_text("");
      output_file_name_->set_text("");
    }
  }

  // Can we do it in one command by putting all controls in one event box?
  if (!force_input_file) {
    input_file_name_->set_sensitive(1);
    input_file_chooser_button_->set_sensitive(1);
  }

  output_file_name_->set_sensitive(1);
  output_file_chooser_button_->set_sensitive(1);
  save_param_button_->set_sensitive(1);
  default_param_button_->set_sensitive(1);
  for (std::vector<Parameter*>::const_iterator i = parameters_.begin();
       i != parameters_.end(); i++) {
    (*i)->get_widget()->set_sensitive(1);
  }
  
  // Make cancel button convert button and enable new conversion
  convert_button_->set_sensitive(1);
  set_convert_button();
}

// Signal handlers
void Frontend::on_message_sig() {
  proceed_message(backend_->get_message());
}
  
void Frontend::on_start_sig() {
  // Enable the cancel button
  convert_button_->set_sensitive(1);
}

void Frontend::on_end_sig() {
  int encoder_ok = (backend_status_ == 1);
  int success = backend_->finish();

  delete backend_;
  backend_ = NULL;
 
  update_gui_upon_end_encoding(encoder_ok, success);
  backend_status_ = 0;
}

// Launch the encoder
void Frontend::on_convert() {
  if (input_file_name_->get_text() == "") {
      return; 	
  }
  backend_status_ = 0;

  update_gui_upon_start_encoding();

  backend_ = new Backend(this);
  backend_->launch(input_file_name_->get_text().c_str(),
                   output_file_name_->get_text().c_str(), parameters_);
}

// Init the file dialogs
// Note that the dialog objects should be reused across
// multiple invocations in order to keep data such as
// the working directory;
// Remeber to hide the dialog after each use
void Frontend::init_file_dialogs() {
  input_file_dialog_ = new Gtk::FileChooserDialog("Please choose a file",
		                     Gtk::FILE_CHOOSER_ACTION_OPEN);
  output_file_dialog_ = new Gtk::FileChooserDialog("Please choose a file",
		                     Gtk::FILE_CHOOSER_ACTION_SAVE);
  
  input_file_dialog_->set_transient_for(*this);
  input_file_dialog_->add_button(Gtk::Stock::CANCEL, Gtk::RESPONSE_CANCEL);
  input_file_dialog_->add_button(Gtk::Stock::OPEN, Gtk::RESPONSE_OK);

  //Add filters, so that only certain file types can be selected:
  Gtk::FileFilter filter_video;
  filter_video.set_name("Video files");
  filter_video.add_mime_type("video/mpeg");
  filter_video.add_mime_type("video/x-mpeg");
  filter_video.add_mime_type("video/quicktime");
  filter_video.add_mime_type("video/x-mpeg2");
  filter_video.add_mime_type("video/x-msvideo");
  filter_video.add_mime_type("video/x-ms-afs");
  filter_video.add_mime_type("video/x-ms-asf");
  filter_video.add_mime_type("video/x-ms-wma");
  filter_video.add_mime_type("video/x-ms-wmv");
  filter_video.add_mime_type("video/x-ms-wmx");
  filter_video.add_mime_type("video/x-ms-wvx");
  input_file_dialog_->add_filter(filter_video);

  Gtk::FileFilter filter_any;
  filter_any.set_name("Any files");
  filter_any.add_pattern("*");
  input_file_dialog_->add_filter(filter_any);

  output_file_dialog_->set_transient_for(*this);
  output_file_dialog_->add_button(Gtk::Stock::CANCEL, Gtk::RESPONSE_CANCEL);
  output_file_dialog_->add_button(Gtk::Stock::OPEN, Gtk::RESPONSE_OK);

  Gtk::FileFilter filter_smv;
  filter_smv.set_name("SMV files");
  filter_smv.add_pattern("*.smv");

  output_file_dialog_->add_filter(filter_smv);
  output_file_dialog_->add_filter(filter_any);
}

void Frontend::update_output_filename(const std::string& input_filename) {
  char *tmp = strdup(input_filename.c_str());
  output_file_name_->set_text(homedir + "/Desktop/" + basename(tmp) + ".smv");
  free(tmp);
}

void Frontend::on_browse_input_file() {
  //Show the dialog and wait for a user response:
  int result = input_file_dialog_->run();

  // Hide the dialog
  input_file_dialog_->hide();

  //Handle the response:
  switch(result)
  {
    case(Gtk::RESPONSE_OK):
    {
      //Notice that this is a std::string, not a Glib::ustring.
      std::string filename = input_file_dialog_->get_filename();
      if (filename != input_file_name_->get_text()) {
	input_file_name_->set_text(filename);
	update_output_filename(filename);
      }
      break;
    }
    case(Gtk::RESPONSE_CANCEL):
      break;
    default:
      break;
  }
}

void Frontend::on_browse_output_file() {
  //Show the dialog and wait for a user response:
  int result = output_file_dialog_->run();

  // Hide the dialog
  output_file_dialog_->hide();

  //Handle the response:
  switch(result)
  {
    case(Gtk::RESPONSE_OK):
      output_file_name_->set_text(output_file_dialog_->get_filename());
      break;
    case(Gtk::RESPONSE_CANCEL):
      break;
    default:
      break;
  }
}

// Signal the encoder
void Frontend::on_cancel() {
  convert_button_->set_sensitive(0);
  backend_->cancel();
}

// The Frontend construction
Frontend::Frontend(BaseObjectType* cobject,
		   const Glib::RefPtr<Gnome::Glade::Xml>& glade_xml)  
  : Gtk::Window(cobject), backend_(NULL) {

  glade_xml->get_widget("InputFilenameEntry", input_file_name_);
  glade_xml->get_widget("OutputFilenameEntry", output_file_name_);
  
  glade_xml->get_widget("InputFileChooserButton", input_file_chooser_button_);
  glade_xml->get_widget("OutputFileChooserButton", output_file_chooser_button_);

  glade_xml->get_widget("SaveParametersButton", save_param_button_);
  glade_xml->get_widget("DefaultParametersButton", default_param_button_);

  glade_xml->get_widget("Progressbar", progress_bar_);

  glade_xml->get_widget("ConvertButton", convert_button_);
  glade_xml->get_widget("QuitButton", quit_button_);

  glade_xml->get_widget("ParamsTable", params_table_);

  progress_bar_->set_text(" ");

  init_file_dialogs();
  init_parameters();
  load_parameters();  

  // Setup buttons
  input_file_chooser_button_->set_label("...");
  output_file_chooser_button_->set_label("...");
  quit_button_->set_image(*(Gtk::manage(
     new Gtk::Image(Gtk::Stock::QUIT, Gtk::ICON_SIZE_BUTTON))));
  quit_button_->set_label(i18n("Quit")); 

  // Do not manage these - they will be destroyed when application quits
  convert_img_ = new Gtk::Image(Gtk::Stock::CONVERT, Gtk::ICON_SIZE_BUTTON);
  cancel_img_ = new Gtk::Image(Gtk::Stock::CANCEL, Gtk::ICON_SIZE_BUTTON);

  set_convert_button();

  // Event handling
  message_dis_.connect(sigc::mem_fun(*this, &Frontend::on_message_sig));
  start_dis_.connect(sigc::mem_fun(*this, &Frontend::on_start_sig));
  end_dis_.connect(sigc::mem_fun(*this, &Frontend::on_end_sig));
  save_param_button_->signal_clicked().connect(
    sigc::mem_fun(*this, &Frontend::save_parameters));
  default_param_button_->signal_clicked().connect(
    sigc::mem_fun(*this, &Frontend::set_default_parameters));
  input_file_chooser_button_->signal_clicked().connect(
       sigc::mem_fun(*this, &Frontend::on_browse_input_file));
  output_file_chooser_button_->signal_clicked().connect(
       sigc::mem_fun(*this, &Frontend::on_browse_output_file));
  quit_button_->signal_clicked().connect(
       sigc::mem_fun(*this, &Frontend::on_quit)); 

  update_gui_after_init();
}

Frontend::~Frontend()
{}

void Frontend::proceed_message(msg_t msg) {
  msg_content_t type = msg.content;

  // Check that we are running the proper backend
  if (backend_status_ == 0) {
     if (msg.tag == hello) {
          backend_status_ = 1;
     } else {
          backend_status_ = -1;
	  backend_->stop();
     }
     return;
  } else if (backend_status_ == -1) return;

  switch(msg.tag) {
  case error:
    progress_bar_->set_text(errstr((err_t)type));
    break;
  case success:
    progress_bar_->set_text(i18n("Finished successfully"));
    break; 
  case interrupted:
    progress_bar_->set_text(i18n("Encoding interrupted"));
    break;
  case phase_start:
  case phase_start_percentage:
    progress_bar_->set_fraction(0);
    progress_bar_->set_text(phasestr((phase_t)type));
    break;
  case phase_end:
    progress_bar_->set_text(" ");
    progress_bar_->set_fraction(0);
    break;
  case alive:
    progress_bar_->pulse();
    break;
  case percentage:
    progress_bar_->set_fraction((double)type / 100);
    break;
  }
  fflush(stdout);
}

const char* Frontend::phasestr(phase_t p) {
  if (p == phase_decode)
    return i18n("Decoding input video");
  if (p == phase_frames)
    return i18n("Extracting frames");
  if (p == phase_tile)
    return i18n("Tiling frames");
  if (p == phase_audio)
    return i18n("Extracting audio");
  if (p == phase_combine)
    return i18n("Generating SMV file");
  return "";
}

const char* Frontend::errstr(err_t e) {
  if (e == err_input_file)
    return i18n("Input video not found");
  if (e == err_output_file_not_writable) 
    return i18n("Output file not writable");
  if (e == err_bad_parameter)
    return i18n("Bad parameter");
  if (e == err_decode || e == err_frames || e == err_tile || e == err_audio ||
      e == err_combine)
    return i18n("Combine failed");
  if (e == err_unexpected)
    return i18n("Unexpected error");
  return "";
}

bool Frontend::on_delete_event(GdkEventAny*)
{
   if (backend_) {
       backend_->cancel();
       backend_->finish();
       delete backend_;
       backend_ = NULL;	
   }
   return 0; 
}

void Frontend::on_quit()
{
   on_delete_event(NULL);
   exit(0);
}

// Init encoder parameter objects
void Frontend::init_parameters() {
  // Create widgets here and call show() on them

  std::vector<std::string> vals; 
  Parameter *param;
  Gtk::Label *title;
  int row;

  vals; vals.clear();
  vals.push_back("96x128"); vals.push_back("127x159");
  param = new EditableTextParameter(i18n("Geometry"), "-g", vals, "96x128"); 
  param->get_widget()->set_tooltip_text(i18n("Select the target video geometry"));
  parameters_.push_back(param); 
    
  vals.clear();
  vals.push_back("10"); vals.push_back("11");
  param = new EditableTextParameter(i18n("Framerate"), "-f", vals, "10");
  param->get_widget()->set_tooltip_text(i18n("Select the target video framerate"));
  parameters_.push_back(param);

  vals.clear();
  vals.push_back("1"); vals.push_back("5");  
  vals.push_back("10"); vals.push_back("11");
  param = new EditableTextParameter(i18n("Tile factor"), "-n", vals, "5");
  param->get_widget()->set_tooltip_text(i18n("Select the number of frames in each jpeg image"));
  parameters_.push_back(param);

  {
    std::vector<std::pair<std::string, std::string> > vals; 
    vals.push_back(std::pair<std::string, std::string>("0°", "-1"));
    vals.push_back(std::pair<std::string, std::string>("90°", "1"));
    vals.push_back(std::pair<std::string, std::string>("180°", "-2"));
    vals.push_back(std::pair<std::string, std::string>("270°", "2"));

    param = new TextParameter(i18n("Rotation"), "-r", vals, "90°");
    param->get_widget()->set_tooltip_text(i18n("Select the rotation angle"));
    parameters_.push_back(param); 
  }

  param = new NumberParameter  (i18n("Image quality"), "-q",  1, 85, 1, 41);
  param->get_widget()->set_tooltip_text(i18n("Select the jpeg image quality"));
  parameters_.push_back(param);

  row = 0;
  for (std::vector<Parameter*>::const_iterator i = parameters_.begin();
       i != parameters_.end(); i++, row++) {
    param = *i;
    if (param -> separate_title()) {
      title = Gtk::manage(new Gtk::Label(param->get_name()));
      title->set_alignment(Gtk::ALIGN_LEFT, Gtk::ALIGN_CENTER);
      title->show();
      params_table_->attach(*title, 0, 1, row, row+1);
      params_table_->attach(*(param->get_widget()), 1, 2, row, row+1);
    } else {
      params_table_->attach(*(param->get_widget()), 0, 2, row, row+1);
    }      
  }
}

// Load the saved encoder parameters from config file
void Frontend::load_parameters() {
  std::ifstream conf(config_file.c_str());

  std::string line, flag, val;
  int n;

  while (conf.good()) {
    getline(conf, line);
    n = line.find_first_of('=');
    flag = line.substr(0, n);
    val = line.substr(n+1);

    for (std::vector<Parameter*>::const_iterator i = parameters_.begin();
	 i != parameters_.end(); i++) {
      if ((*i)->get_flag() == flag) {
	(*i)->set_val(val);
	break;
      }
    }
  }
  
}

// Save encoder parameters
void Frontend::save_parameters() {
  std::ofstream conf(config_file.c_str());

  for (std::vector<Parameter*>::const_iterator i = parameters_.begin();
       i != parameters_.end(); i++) {
    conf << (*i)->get_flag() << '=' << (*i)->get_val() << std::endl;
  }

  conf.close();
}

// Reset defaults
void Frontend::set_default_parameters() {
  for (std::vector<Parameter*>::const_iterator i = parameters_.begin();
       i != parameters_.end(); i++) {
    (*i)->reset();
  }
  save_parameters();
}

// Encoder parameter objects
EditableTextParameter::EditableTextParameter(const std::string& name, 
					     const std::string& flag,
		     const std::vector<std::string>& all_vals,
		     const std::string& initial_value) :
  name_(name), flag_(flag), 
  all_vals_(all_vals), initial_value_(initial_value) {
  widget_ = Gtk::manage(new Gtk::ComboBoxEntryText());
  for (std::vector<std::string>::const_iterator i = 
	 all_vals.begin(); i != all_vals.end(); i++) {
    ((Gtk::ComboBoxEntryText*)widget_)->append_text(*i);
  }
  set_val(initial_value_);
  widget_->show();
}

std::string EditableTextParameter::get_val() {
  return ((Gtk::ComboBoxEntryText*)widget_)->get_entry()->get_text();
}

std::string EditableTextParameter::get_flag_and_val() {
  return flag_ + " " + ((Gtk::ComboBoxEntryText*)widget_)->get_entry()->get_text();
}

void EditableTextParameter::set_val(const std::string& val) {
  ((Gtk::ComboBoxEntryText*)widget_)->get_entry()->set_text(val);
}

void EditableTextParameter::reset() {
  set_val(initial_value_); 
}

TextParameter::TextParameter(const std::string& name, const std::string& flag,
	  const std::vector<std::pair<std::string, std::string> >& vals,
          const std::string& initial_value) :
  name_(name), flag_(flag), initial_value_(initial_value)
{
  widget_ = Gtk::manage(new Gtk::ComboBoxText());
  for (std::vector<std::pair<std::string, std::string> >::const_iterator i = 
	 vals.begin(); i != vals.end(); i++) {
    ((Gtk::ComboBoxText*)widget_)->append_text(i->first);
    all_vals_[i->first] = i->second;
  }
  set_val(initial_value_);
  widget_->show();
}

std::string TextParameter::get_flag_and_val() {
  return flag_ + " " + all_vals_[((Gtk::ComboBoxText*)widget_)->get_active_text()];
}

std::string TextParameter::get_val() {
  return ((Gtk::ComboBoxText*)widget_)->get_active_text();
}

void TextParameter::set_val(const std::string& val) {
    if (all_vals_.find(val) != all_vals_.end())
      ((Gtk::ComboBoxText*)widget_)->set_active_text(val);
}

void TextParameter::reset() {
  set_val(initial_value_); 
}
 
NumberParameter::NumberParameter(const std::string& name, const std::string& flag,
	     double min, double max, double step, double initial_value) :
  name_(name), flag_(flag),
  min_(min), max_(max), step_(step), initial_value_(initial_value)
{
  widget_ = Gtk::manage(new Gtk::HScale(min, max, step));
//  ((Gtk::HScale*)widget_)->set_increments(1, 10);
//  ((Gtk::HScale*)widget_)->set_range(min, max);
  ((Gtk::HScale*)widget_)->set_value(initial_value);
  widget_->show();
}

std::string NumberParameter::get_flag_and_val() {
  std::stringstream ss;
  ss << flag_ << " " << get_val();
  return ss.str();
}

std::string NumberParameter::get_val() {
  std::stringstream ss;
  ss << (((Gtk::HScale*)widget_)-> get_value());
  return ss.str();
}

void NumberParameter::reset() {
  ((Gtk::HScale*)widget_)-> set_value(initial_value_);
}
  
void NumberParameter::set_val(const std::string& val) {
  double x;
  std::stringstream ss(val);
  if ((ss >> x) && x>= min_ && x<=max_) {
    ((Gtk::HScale*)widget_)-> set_value(x);
  }
}

// Determine the installation directory
// The following C code should be placed in a separate file
// I found it on the internet

#ifndef PATH_MAX
  #define PATH_MAX 4096
#endif

/* Check if file exists */
static int checkifexists(const char *name) {
  struct stat s_buf;
  return (stat(name, &s_buf) != -1);
}

/* int checkifexecutable(const char *filename)
 * 
 * Return non-zero if the name is an executable file, and
 * zero if it is not executable, or if it does not exist.
 */

int checkifexecutable(const char *filename)
{
     int result;
     struct stat statinfo;
     
     result = stat(filename, &statinfo);
     if (result < 0) return 0;
     if (!S_ISREG(statinfo.st_mode)) return 0;

     if (statinfo.st_uid == geteuid()) return statinfo.st_mode & S_IXUSR;
     if (statinfo.st_gid == getegid()) return statinfo.st_mode & S_IXGRP;
     return statinfo.st_mode & S_IXOTH;
}


/* int findpathof(char *pth, const char *exe)
 *
 * Find executable by searching the PATH environment variable.
 *
 * const char *exe - executable name to search for.
 *       char *pth - the path found is stored here, space
 *                   needs to be available.
 *
 * If a path is found, returns non-zero, and the path is stored
 * in pth.  If exe is not found returns 0, with pth undefined.
 */

int findpathof(char *canonic, const char *exe)
{
     char *searchpath;
     char *beg, *end;
     char pth[PATH_MAX + 1]; 
     int stop, found;
     int len;

     if (strchr(exe, '/') != NULL) {
	  if (realpath(exe, canonic) == NULL) return 0;
	  return  checkifexecutable(canonic);
     }

     searchpath = getenv("PATH");
     if (searchpath == NULL) return 0;
     if (strlen(searchpath) <= 0) return 0;

     beg = searchpath;
     stop = 0; found = 0;
     do {
	  end = strchr(beg, ':');
	  if (end == NULL) {
	       stop = 1;
	       strncpy(pth, beg, PATH_MAX);
	       len = strlen(pth);
	  } else {
	       strncpy(pth, beg, end - beg);
	       pth[end - beg] = '\0';
	       len = end - beg;
	  }
	  if (pth[len - 1] != '/') strncat(pth, "/", 1);
	  strncat(pth, exe, PATH_MAX - len);
	  found = checkifexecutable(pth);
	  if (!stop) beg = end + 1;
     } while (!stop && !found);
     return (realpath(pth, canonic) != NULL);
}

int main (int argc, char ** argv) {
  std::string install_dir, glade_filename;

  // initialize gettext 
  setlocale (LC_ALL, "");
  bindtextdomain (PACKAGE, LOCALEDIR);
  textdomain (PACKAGE);

  // Invocation from a file manager popup menu
  if (argc > 1) {
    force_input_file = strdup(argv[1]);
  }

  char canonic_exe[PATH_MAX + 1];
  if (!findpathof(canonic_exe, argv[0])) {
      printf("DEBUG Failed to determine installation directory\n");
      exit(-1);	
  }
  
  install_dir = strdup(dirname(strdup(canonic_exe)));
  printf("DEBUG install bin dir: %s\n", install_dir.c_str());  

  glade_filename = install_dir + "/smv_gui.glade";
  if (!checkifexists(glade_filename.c_str())) {
    printf("DEBUG missing: %s\n", glade_filename.c_str());
    glade_filename = install_dir + "/../share/smv_gui/smv_gui.glade";
  }
  
  if (!checkifexists(glade_filename.c_str())) {
      printf("DEBUG missing: %s\n", glade_filename.c_str());  
      exit(-1);
  } else {
    printf("DEBUG found: %s\n", glade_filename.c_str());
  }

  Glib::thread_init();
  Glib::RefPtr<Gnome::Glade::Xml> main_window;
  Gtk::Main kit(argc, argv);

  // load the interface
  main_window = Gnome::Glade::Xml::create(glade_filename);

  // Get the main window
  Frontend* window;
  main_window->get_widget_derived("MainWindow", window);

  // start the event loop
  Gtk::Main::run( *window );

  return 0;
}
